CREATE OR REPLACE PACKAGE BODY MCL_COMPRESS_CALC AS
/*
* Package:    MCL_COMPRESS_CALC
* Date:       29-04-2011
* Author:     M.Lambrechts <marcel.lambrechts@gmail.com>
* Version:    1.1
*
* Changes:
* Date			Who		Version		What
* =============================================================================
* 29-04-2011	ML		1.0			Initial creation
* 10-05-2011	ML		1.0			Released
* 11-05-2011	ML		1.1			Updated PROCESS_WORKLIST procedure using
*									work_id column in worklist table
*									Updated SETUP_TABLES procedure to add
*									work_id column to worklist table and a
*									primary key to make it unique
*									Added ADD_TO_WORKLIST procedure for adding
*									objects to worklist table									
* =============================================================================
*
*/

PROCEDURE GET_RATIO_FOR_TABLE
  (owner IN VARCHAR2, tablename IN VARCHAR2, partname IN VARCHAR2 DEFAULT NULL,
   compress_type IN VARCHAR2 DEFAULT 'DIRECT_LOAD', sample_pct IN NUMBER DEFAULT NULL,
   samp_tsname IN VARCHAR2 DEFAULT NULL, show_samp_usage IN BOOLEAN DEFAULT FALSE,
   store_stats IN BOOLEAN DEFAULT FALSE)
AS
  lv_db_version			VARCHAR2(20);
  lv_db_compat			VARCHAR2(20);
  lv_count				NUMBER;
  lv_size_div			NUMBER;
  lv_src_size			NUMBER;
  lv_blocks				NUMBER;
  lv_samp_usage			NUMBER;
  lv_blocks_utab		NUMBER;
  lv_blocks_ctab		NUMBER;
  lv_rblocks_utab		NUMBER;
  lv_rblocks_ctab		NUMBER;
  lv_comp_ratio			NUMBER;
  lv_sec_epoch			NUMBER;
  lv_samp_pct			NUMBER;
  lv_src_tsblocksize	NUMBER;
  lv_samp_tsblocksize	NUMBER;
  lv_size_div_unit		CHAR(2);
  lv_compression		VARCHAR2(10);
  lv_tmp_tab            VARCHAR2(50);
  lv_samp_syn			VARCHAR2(50);
  lv_samp_utab			VARCHAR2(50);
  lv_samp_ctab			VARCHAR2(50);
  lv_comp_stmt			VARCHAR2(30);
  lv_src_tsname			VARCHAR2(30);
  lv_samp_tsname		VARCHAR2(30);
  lv_stmt				VARCHAR2(2000);
BEGIN

-- Get database version and compatible setting
--SELECT version INTO lv_db_version FROM v$instance;
SELECT value INTO lv_db_compat FROM v$parameter WHERE name='compatible';

SELECT (SYSDATE - CAST('01-JAN-1970' AS DATE)) * 86400 INTO lv_sec_epoch FROM dual;

lv_sec_epoch := ROUND(lv_sec_epoch);
lv_samp_syn  := 'SYN_MCLC_'||lv_sec_epoch;
lv_samp_utab := 'MCLCU'||lv_sec_epoch;
lv_samp_ctab := 'MCLCC'||lv_sec_epoch;
lv_tmp_tab   := 'MCLCT'||lv_sec_epoch;

/*
* Run various checks on arguments and set default values based on these arguments
*/
-- Check if specified table exists, if not exit.
SELECT COUNT(1) INTO lv_count FROM all_objects WHERE owner=owner AND object_name=tablename AND object_type = 'TABLE';
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'Table does not exist or you don''t have the correct privileges');
END IF;

-- Check if specified (if any) partition exists
IF (NOT partname IS NULL) THEN
  SELECT COUNT(1) INTO lv_count FROM all_tab_partitions WHERE owner=owner AND table_name=tablename AND partition_name=partname;
  IF (lv_count = 0) THEN
    RAISE_APPLICATION_ERROR(-20001,'Table partition does not exist or you don''t have the correct privileges to see it');
  END IF;
  -- Make sure the partition is not already compressed
  SELECT compression INTO lv_compression FROM all_tab_partitions WHERE table_owner=owner AND table_name=tablename AND partition_name=partname;
  IF (lv_compression = 'ENABLED') THEN
    RAISE_APPLICATION_ERROR(-20001,'Table partition is already compressed');
  END IF;
ELSE
  SELECT compression INTO lv_compression FROM all_tables WHERE owner=owner AND table_name=tablename;
  IF (lv_compression = 'ENABLED') THEN
    RAISE_APPLICATION_ERROR(-20001,'Table is already compressed');
  END IF;
END IF;

-- Check if specified compress_type is correct, if not exit.
IF ((compress_type != 'DIRECT_LOAD') AND (compress_type != 'OLTP')) THEN
  RAISE_APPLICATION_ERROR(-20001,'Specified compression type not supported!');
ELSE
  -- 3. check if specified compress_type is supported by the database version, if not exit.
  IF ((compress_type = 'OLTP') AND (CAST(SUBSTR(lv_db_compat,1,INSTR(lv_db_compat,'.')-1) AS NUMBER) < 11)) THEN
    RAISE_APPLICATION_ERROR(-20001,'OLTP Table compression not supported for database versions before 11g release 1');
  END IF;
END IF;

-- Get the size (in blocks and bytes) of the table or table partition
IF (NOT partname IS NULL) THEN
  SELECT SUM(blocks), SUM(bytes) INTO lv_blocks, lv_src_size FROM dba_segments WHERE owner=owner AND segment_name=tablename AND partition_name=partname;
ELSE
  SELECT SUM(blocks), SUM(bytes) INTO lv_blocks, lv_src_size FROM dba_segments WHERE owner=owner AND segment_name=tablename;
END IF;

-- Check the specified sample percentage (< 100) and if null, determine a default percentage based on 1000 blocks
IF (NOT sample_pct IS NULL) THEN
  IF ((NOT sample_pct > 0.000001) AND (NOT sample_pct < 100)) THEN
    RAISE_APPLICATION_ERROR(-20001,'Sample percentage should be between 0.000001 and 100');
  ELSE
    lv_samp_pct := sample_pct;
  END IF;
ELSE
  -- Determine percentage to use based on the default maximum number of sample blocks
  SELECT (LEAST(DEF_SAMP_BLOCKS,lv_blocks)/lv_blocks)*99.999 INTO lv_samp_pct FROM dual;
END IF;

-- Get information about the tablespace where the source object is stored and the tablespace specified (if any) for storing the sample table
IF (NOT partname IS NULL) THEN
  SELECT tablespace_name INTO lv_src_tsname FROM dba_segments WHERE owner=owner AND segment_name=tablename AND partition_name=partname;
ELSE
  SELECT tablespace_name INTO lv_src_tsname FROM dba_segments WHERE owner=owner AND segment_name=tablename AND rownum=1;
END IF;
SELECT block_size INTO lv_src_tsblocksize FROM dba_tablespaces WHERE tablespace_name=lv_src_tsname;

IF (samp_tsname IS NULL) THEN
  BEGIN
    lv_stmt := 'SELECT tablespace_name FROM dba_segments WHERE segment_name='''||lv_tmp_tab||'''';
    EXECUTE IMMEDIATE 'CREATE TABLE '||lv_tmp_tab||' (id NUMBER)';
    EXECUTE IMMEDIATE lv_stmt INTO lv_samp_tsname;
    EXECUTE IMMEDIATE 'DROP TABLE '||lv_tmp_tab||' PURGE';
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to create/drop table to get information about the default tablespace');
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
  END;
ELSE
  lv_samp_tsname := samp_tsname;
END IF;
SELECT block_size INTO lv_samp_tsblocksize FROM dba_tablespaces WHERE tablespace_name=lv_samp_tsname;

-- Generate a warning if tablespace blocksize between tablespace of source object and sample tables are different
IF (lv_samp_tsblocksize <> lv_src_tsblocksize) THEN
  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - NOTE] Blocksize of tablespaces of source object ('||lv_src_tsblocksize||
  ') and sample tables ('||lv_samp_tsblocksize||') differ');
END IF;

/*
* Now only show the rough size used by the sample tables (show_samp_usage = TRUE) or
* let's start and actually calculate the compression factor for the specified table/mview or table partition
*/
IF (show_samp_usage) THEN
  -- Calculate space usage using formula: ((sample_pct/100) * segment_size) + (sample_pct/100) * segment_size)/2.5
  lv_samp_usage := ROUND(((lv_samp_pct/100) * lv_src_size) + (((lv_samp_pct/100) * lv_src_size)/2.5));
  -- Determine value for dividing size of sample usage
  IF (lv_samp_usage > SIZE_GB) THEN
	lv_size_div := SIZE_GB;
	lv_size_div_unit := 'GB';
  ELSE
    lv_size_div := SIZE_MB;
	lv_size_div_unit := 'MB';
  END IF;
  DBMS_OUTPUT.PUT_LINE('Estimated space used for determination of compression factor (sample size: '||ROUND(lv_samp_pct,1)||'%): '||
  ROUND(lv_samp_usage/lv_size_div,1)||' '||lv_size_div_unit);
-- 5. if show_temp_usage is FALSE (default) continue
ELSE
  -- Get the correct compress statement, depending on the version of the Oracle database
  IF (compress_type = 'DIRECT_LOAD') THEN
    IF (CAST(SUBSTR(lv_db_compat,1,INSTR(lv_db_compat,'.')-1) AS NUMBER) < 11) THEN
	  lv_comp_stmt := 'COMPRESS';
	ELSE
	  lv_comp_stmt := 'COMPRESS FOR DIRECT_LOAD OPERATIONS';
	END IF;
  ELSIF (compress_type = 'OLTP') THEN
    IF (SUBSTR(lv_db_compat,1,INSTR(lv_db_compat,'.',1,2)-1) = '11.1') THEN
	  lv_comp_stmt := 'COMPRESS FOR ALL OPERATIONS';
	ELSE
	  -- Compress for OLTP replaces FOR ALL OPERATIONS as of 11.2
	  lv_comp_stmt := 'COMPRESS FOR OLTP';
	END IF;
  END IF;

  -- Create synonym for original table
  BEGIN
    lv_stmt := 'CREATE OR REPLACE SYNONYM '||lv_samp_syn||' FOR '||owner||'.'||tablename;
    EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to create a synonym for the target table/materialized view');
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Failing statement: '||lv_stmt);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
	  RETURN;
  END;
  
  -- Create a uncompressed sample table based on the original table to calculate the compression ratio for
  BEGIN
    IF (samp_tsname IS NULL) THEN
	  lv_stmt := 'CREATE TABLE '||lv_samp_utab||' NOCOMPRESS AS SELECT /*+ full('||lv_samp_syn||')*/ * FROM '||lv_samp_syn;
	ELSE
	  lv_stmt := 'CREATE TABLE '||lv_samp_utab||' NOCOMPRESS TABLESPACE '||samp_tsname||' AS SELECT /*+ FULL('||lv_samp_syn||')*/ * FROM '||lv_samp_syn;
    END IF;
	
	IF (NOT partname IS NULL) THEN
	  lv_stmt := lv_stmt || ' PARTITION ('||partname||')';
	END IF;
	lv_stmt := lv_stmt || ' SAMPLE BLOCK('||lv_samp_pct||',1)';
	
	EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to create an uncompressed sample table.');
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Failing statement: '||lv_stmt);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
	  RETURN;
  END;
  
  -- Create a compressed table from the uncompressed sample table
  BEGIN
    IF (samp_tsname IS NULL) THEN
	  lv_stmt := 'CREATE TABLE '||lv_samp_ctab||' '||lv_comp_stmt||' AS SELECT * FROM '||lv_samp_utab;
	ELSE
	  lv_stmt := 'CREATE TABLE '||lv_samp_ctab||' '||lv_comp_stmt||' TABLESPACE '||samp_tsname||' AS SELECT * FROM '||lv_samp_utab;
	END IF;
    EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to create an compressed sample table based on the uncompressed sample table.');
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Cleanup uncompressed sample table before quiting.');
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Failing statement: '||lv_stmt);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
	  EXECUTE IMMEDIATE 'DROP TABLE '||lv_samp_utab||' PURGE';
	  RETURN;
  END;
  
  -- Get the number of blocks and rows per block for the uncompressed and compressed sample table
  lv_stmt := 'SELECT COUNT(DISTINCT(dbms_rowid.rowid_block_number(rowid))) FROM '||lv_samp_utab;
  EXECUTE IMMEDIATE lv_stmt INTO lv_blocks_utab;
  
  lv_stmt := 'SELECT COUNT(DISTINCT(dbms_rowid.rowid_block_number(rowid))) FROM '||lv_samp_ctab;
  EXECUTE IMMEDIATE lv_stmt INTO lv_blocks_ctab;
  
  lv_stmt := 'SELECT AVG(cnt) FROM (SELECT COUNT(*) cnt FROM '||lv_samp_utab||' GROUP BY dbms_rowid.rowid_block_number(rowid))';
  EXECUTE IMMEDIATE lv_stmt INTO lv_rblocks_utab;
  
  lv_stmt := 'SELECT AVG(cnt) FROM (SELECT COUNT(*) cnt FROM '||lv_samp_ctab||' GROUP BY dbms_rowid.rowid_block_number(rowid))';
  EXECUTE IMMEDIATE lv_stmt INTO lv_rblocks_ctab;

  -- Calculate the compression ratio based on the number of blocks
  lv_comp_ratio := lv_blocks_utab/lv_blocks_ctab;
  
  -- Determine value for dividing size of source segment
  IF (lv_src_size > SIZE_GB) THEN
	lv_size_div := SIZE_GB;
	lv_size_div_unit := 'GB';
  ELSE
    lv_size_div := SIZE_MB;
	lv_size_div_unit := 'MB';
  END IF;
  
  IF (NOT partname IS NULL) THEN
    DBMS_OUTPUT.PUT_LINE('Compression ratio for table partition '||partname||' table ('||owner||'.'||tablename||') is: '||ROUND(lv_comp_ratio,1));
    DBMS_OUTPUT.PUT_LINE('(Size of table partition '||partname||' table ('||owner||'.'||tablename||') would be reduced from '||
	ROUND(lv_src_size/lv_size_div,1)||' '||lv_size_div_unit||' to '||ROUND((lv_src_size/lv_comp_ratio)/lv_size_div,1)||' '||lv_size_div_unit||')');  
  ELSE
    DBMS_OUTPUT.PUT_LINE('Compression ratio for table '||owner||'.'||tablename||' is: '||ROUND(lv_comp_ratio,1));
    DBMS_OUTPUT.PUT_LINE('(Size of table/mview '||owner||'.'||tablename||' could be reduced from '||
	ROUND(lv_src_size/lv_size_div,1)||' '||lv_size_div_unit||' to '||ROUND((lv_src_size/lv_comp_ratio)/lv_size_div,1)||' '||lv_size_div_unit||')');
  END IF;
  DBMS_OUTPUT.PUT_LINE('.');
  DBMS_OUTPUT.PUT_LINE('Source object information');
  DBMS_OUTPUT.PUT_LINE('==========================================================');
  DBMS_OUTPUT.PUT_LINE('Tablespace source object:             '||lv_src_tsname);
  DBMS_OUTPUT.PUT_LINE('Tablespace source object block size:  '||lv_src_tsblocksize);
  DBMS_OUTPUT.PUT_LINE('.');
  DBMS_OUTPUT.PUT_LINE('Sample information');
  DBMS_OUTPUT.PUT_LINE('==========================================================');
  DBMS_OUTPUT.PUT_LINE('Tablespace sample table               '||lv_samp_tsname);
  DBMS_OUTPUT.PUT_LINE('Tablespace sample table block size:   '||lv_samp_tsblocksize);
  DBMS_OUTPUT.PUT_LINE('Sample size (%):                      '||ROUND(lv_samp_pct));
  DBMS_OUTPUT.PUT_LINE('Blocks sample table uncompressed:     '||lv_blocks_utab);
  DBMS_OUTPUT.PUT_LINE('Blocks sample table compressed:       '||lv_blocks_ctab);
  DBMS_OUTPUT.PUT_LINE('Rows/block sample table uncompressed: '||ROUND(lv_rblocks_utab));
  DBMS_OUTPUT.PUT_LINE('Rows/block sample table compressed:   '||ROUND(lv_rblocks_ctab));

  IF (store_stats) THEN  
    IF (NOT partname IS NULL) THEN
      STORE_COMPRESS_STATS('PARTITION',owner,tablename,partname,compress_type,lv_samp_pct,0,lv_comp_ratio,lv_src_size,lv_src_size/lv_comp_ratio);
    ELSE
      STORE_COMPRESS_STATS('TABLE',owner,tablename,NULL,compress_type,lv_samp_pct,0,lv_comp_ratio,lv_src_size,lv_src_size/lv_comp_ratio);
    END IF;
  END IF;
  
  -- Remove sample tables
  BEGIN
    lv_stmt := 'DROP TABLE '||lv_samp_utab||' PURGE';
	EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to drop uncompressed sample table '||lv_samp_utab);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
  END;
  
  BEGIN
    lv_stmt := 'DROP TABLE '||lv_samp_ctab||' PURGE';
	EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to drop compressed sample table '||lv_samp_ctab);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
  END;
  
  -- Remove synonym to original table
  BEGIN
    lv_stmt := 'DROP SYNONYM '||lv_samp_syn;
	EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] Unable to drop synonym '||lv_samp_syn);
	  DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
  END;
END IF;

EXCEPTION
	WHEN OTHERS THEN
		DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_TABLE - ERROR] General error');
		RAISE_APPLICATION_ERROR(-20002,'[GET_RATIO_FOR_TABLE - ERROR] General error: '||SUBSTR(SQLERRM, 1, 100));
END;


PROCEDURE GET_RATIO_FOR_INDEX
  (owner IN VARCHAR2, indexname IN VARCHAR2, store_stats IN BOOLEAN DEFAULT FALSE)
AS
  lv_count				NUMBER;
  lv_size_div			NUMBER;
  lv_src_size			NUMBER;
  lv_blocks				NUMBER;
  lv_opt_pref_length	NUMBER;
  lv_pct_saved			NUMBER;
  lv_comp_ratio			NUMBER;
  lv_size_div_unit		CHAR(2);
  lv_uniqueness			VARCHAR2(10);
  lv_index_type			VARCHAR2(30);
  lv_compression		VARCHAR2(10);
  lv_partitioned		VARCHAR2(3);
  lv_stmt				VARCHAR2(2000);
BEGIN

-- Make sure index exists before we continue
SELECT COUNT(1) INTO lv_count FROM all_objects WHERE owner=owner AND object_name=indexname AND object_type='INDEX';
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'Index does not exist or you don''t have the correct privileges');
END IF;

-- Get information about the index
SELECT index_type, uniqueness, compression, partitioned INTO lv_index_type, lv_uniqueness, lv_compression, lv_partitioned
FROM all_indexes WHERE owner=owner AND index_name=indexname;

-- Perform various checks on the specified index to make sure compression is supported
IF (NOT lv_compression = 'DISABLED') THEN
  RAISE_APPLICATION_ERROR(-20001,'Index is already compressed');
END IF;

IF (lv_partitioned = 'YES') THEN
  RAISE_APPLICATION_ERROR(-20001,'Index is partitioned. Index key compression not supported');
END IF;

IF (NOT lv_index_type = 'NORMAL') THEN
  RAISE_APPLICATION_ERROR(-20001,'Not a B-Tree index. Index key compression not supported');
END IF;

-- Check if the index is non-unique or an unique index with at least 2 columns
IF (lv_uniqueness = 'UNIQUE') THEN
	SELECT COUNT(1) INTO lv_count FROM all_ind_columns WHERE index_owner=owner AND index_name=indexname;
	IF (lv_count <= 2) THEN
	  RAISE_APPLICATION_ERROR(-20001,'Index is a unique index on a single column. Compression is not supported');
	END IF;
END IF;

-- Run a ANALYZE INDEX .... VALIDATE STRUCTURE
BEGIN
  lv_stmt := 'ANALYZE INDEX '||owner||'.'||indexname||' VALIDATE STRUCTURE';
  EXECUTE IMMEDIATE lv_stmt;
EXCEPTION
  WHEN OTHERS THEN
    DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_INDEX - ERROR] Unable to perform an ANALYZE with VALIDATE STRUCTURE');
    DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
	RAISE_APPLICATION_ERROR(-20001,'[GET_RATIO_FOR_INDEX - ERROR] Premature exit');
END;

-- Check if information about the index is available
SELECT COUNT(1) INTO lv_count FROM index_stats WHERE name=indexname;
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'[GET_RATIO_FOR_INDEX - ERROR] No information about the index available in index_stats table');
END IF;

-- Results of VALIDATE STRUCTURE will be stored in INDEX_STATS table, so get it from there
SELECT blocks, opt_cmpr_count, opt_cmpr_pctsave INTO lv_blocks, lv_opt_pref_length, lv_pct_saved FROM index_stats WHERE name=indexname;

-- Get size (in bytes) of index
SELECT bytes INTO lv_src_size FROM dba_segments WHERE owner=owner AND segment_name=indexname AND segment_type='INDEX';

-- Calculate the compression ratio
lv_comp_ratio := 100/(100-lv_pct_saved);

-- Determine value for dividing size of source segment
IF (lv_src_size > SIZE_GB) THEN
  lv_size_div := SIZE_GB;
  lv_size_div_unit := 'GB';
ELSE
  lv_size_div := SIZE_MB;
  lv_size_div_unit := 'MB';
END IF;

IF (lv_comp_ratio = 1) THEN
  DBMS_OUTPUT.PUT_LINE('Compression of this index would not result in any storage reduction');
  IF (store_stats) THEN  
    STORE_COMPRESS_STATS('INDEX',owner,indexname,NULL,'IKC',0,0,1,lv_src_size,lv_src_size);
  END IF;
ELSE
  DBMS_OUTPUT.PUT_LINE('Compression ratio for index '||owner||'.'||indexname||' with a prefix length of '||lv_opt_pref_length||' key(s) is: '||ROUND(lv_comp_ratio,1));
  DBMS_OUTPUT.PUT_LINE('(Size of index '||owner||'.'||indexname||' would be reduced from '||
  ROUND(lv_src_size/lv_size_div,1)||' '||lv_size_div_unit||' to '||ROUND((lv_src_size-((lv_pct_saved/100)*lv_src_size))/lv_size_div,1)||' '||lv_size_div_unit||')');
  IF (store_stats) THEN
    STORE_COMPRESS_STATS('INDEX',owner,indexname,NULL,'INDEX KEY',0,lv_opt_pref_length,lv_comp_ratio,lv_src_size,lv_src_size-((lv_pct_saved/100)*lv_src_size));
  END IF;
END IF;

EXCEPTION
	WHEN OTHERS THEN
		DBMS_OUTPUT.PUT_LINE('[GET_RATIO_FOR_INDEX - ERROR] General error');
		RAISE_APPLICATION_ERROR(-20002,'[GET_RATIO_FOR_INDEX - ERROR] General error: '||SUBSTR(SQLERRM, 1, 100));
END;

PROCEDURE PROCESS_WORKLIST
AS
  TYPE WLCurType IS REF CURSOR;
  lv_wlcur 			WLCurType; 
  lv_count			NUMBER;
  lv_id				NUMBER;
  lv_error			BOOLEAN;
  lv_stmt			VARCHAR2(2000);
  lv_type			VARCHAR2(20);
  lv_owner			VARCHAR2(30);
  lv_objname		VARCHAR2(30);
  lv_partition		VARCHAR2(30);
  lv_comp_type		VARCHAR2(15);
  lv_samp_pct		NUMBER;
  lv_samp_tsname	VARCHAR2(30);
BEGIN

-- Make sure the worklist table exists
SELECT COUNT(1) INTO lv_count FROM user_tables WHERE table_name=WORKLIST_TABLE;
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'[PROCESS_WORKLIST - ERROR] Worklist table '||WORKLIST_TABLE||' does not exist'); 
END IF;

lv_stmt := 'SELECT work_id, obj_type, obj_owner, obj_name, obj_partition, compress_type, sample_pct, sample_ts FROM '||WORKLIST_TABLE||' WHERE status IS NULL';
OPEN lv_wlcur FOR lv_stmt;
LOOP
  FETCH lv_wlcur into lv_id, lv_type, lv_owner, lv_objname, lv_partition, lv_comp_type, lv_samp_pct, lv_samp_tsname;
  EXIT WHEN lv_wlcur%NOTFOUND;
  lv_type := UPPER(lv_type);
  lv_error := FALSE;
  IF (lv_type IN ('INDEX','TABLE','PARTITION')) THEN
    -- Set status for object to 'IN PROGRESS'
    lv_stmt := 'UPDATE '||WORKLIST_TABLE||' SET start_dt=SYSDATE, status=''IN PROGRESS'' WHERE work_id='||lv_id;
    EXECUTE IMMEDIATE lv_stmt;
	COMMIT;
	
    -- Depending on the specified object type (TABLE, PARTITION or INDEX) execute correct procedure for getting the compression statistics  
    IF (lv_type = 'TABLE' OR lv_type = 'PARTITION') THEN
	  BEGIN
        GET_RATIO_FOR_TABLE(lv_owner,lv_objname,lv_partition,lv_comp_type,lv_samp_pct,lv_samp_tsname,FALSE,TRUE);
	  EXCEPTION
	    WHEN OTHERS THEN
		  lv_stmt := 'UPDATE '||WORKLIST_TABLE||' SET finish_dt=SYSDATE, status=''ERROR'', emsg='''||SUBSTR(SQLERRM, 1, 100)||''' WHERE work_id='||lv_id;
          EXECUTE IMMEDIATE lv_stmt;
	      COMMIT;
		  lv_error := TRUE;
	  END;
    ELSIF (lv_type = 'INDEX') THEN
	  BEGIN
        GET_RATIO_FOR_INDEX(lv_owner,lv_objname,TRUE);
	  EXCEPTION
	    WHEN OTHERS THEN
		  lv_stmt := 'UPDATE '||WORKLIST_TABLE||' SET finish_dt=SYSDATE, status=''ERROR'', emsg='''||SUBSTR(SQLERRM, 1, 100)||''' WHERE work_id='||lv_id;
          EXECUTE IMMEDIATE lv_stmt;
	      COMMIT;
		  lv_error := TRUE;
	  END;
    END IF;
	
	IF (NOT lv_error) THEN
	  -- Set status for object to 'IN PROGRESS'
      lv_stmt := 'UPDATE '||WORKLIST_TABLE||' SET finish_dt=SYSDATE, status=''DONE'' WHERE work_id='||lv_id;
      EXECUTE IMMEDIATE lv_stmt;
	  COMMIT;
	END IF;
  ELSE
    DBMS_OUTPUT.PUT_LINE('[PROCESS_WORKLIST - WARNING] object type '||lv_type||' not a supported object type (supported types are INDEX, TABLE and PARTITION)');
  END IF;  
END LOOP;
CLOSE lv_wlcur; 

EXCEPTION
	WHEN OTHERS THEN
	  RAISE_APPLICATION_ERROR(-20002,'[PROCESS_WORKLIST - ERROR] General error: '||SUBSTR(SQLERRM, 1, 100));
END;

PROCEDURE ADD_TO_WORKLIST
  (obj_type IN VARCHAR2, obj_owner IN VARCHAR2, obj_name IN VARCHAR2, obj_partition IN VARCHAR2 DEFAULT NULL, compress_type IN VARCHAR2 DEFAULT NULL,
   sample_pct IN NUMBER DEFAULT NULL, sample_ts IN VARCHAR2 DEFAULT NULL)
AS
  lv_count			NUMBER;
  lv_newid			NUMBER;
  lv_collist		VARCHAR2(200);
  lv_vallist		VARCHAR2(300);
  lv_stmt			VARCHAR2(2000);
BEGIN

-- Make sure the worklist table exists
SELECT COUNT(1) INTO lv_count FROM user_tables WHERE table_name=WORKLIST_TABLE;
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'[ADD_TO_WORKLIST - ERROR] Compression worklist table '||WORKLIST_TABLE||' does not exist');
END IF; 

-- Get a new id number (max + 1)
lv_stmt := 'SELECT NVL(MAX(work_id)+1,1) FROM '||WORKLIST_TABLE;
EXECUTE IMMEDIATE lv_stmt INTO lv_newid;

-- Mandatory columns and values
lv_collist := '(work_id,obj_type,obj_owner,obj_name';
lv_vallist := '('||lv_newid||','''||obj_type||''','''||obj_owner||''','''||obj_name||'''';

IF (NOT obj_partition IS NULL) THEN
  lv_collist := lv_collist||',obj_partition';
  lv_vallist := lv_vallist||','''||obj_partition||'''';
END IF;
IF (NOT compress_type IS NULL) THEN
  lv_collist := lv_collist||',compress_type';
  lv_vallist := lv_vallist||','''||compress_type||'''';
END IF;
IF (NOT sample_pct IS NULL) THEN
  lv_collist := lv_collist||',sample_pct';
  lv_vallist := lv_vallist||','||sample_pct;
END IF;
IF (NOT sample_ts IS NULL) THEN
  lv_collist := lv_collist||',sample_ts';
  lv_vallist := lv_vallist||','''||sample_ts||'''';
END IF;

-- Close the column list and values list
lv_collist := lv_collist||')';
lv_vallist := lv_vallist||')';

-- Insert object into worklist table
lv_stmt := 'INSERT INTO '||WORKLIST_TABLE||' '||lv_collist||' VALUES '||lv_vallist;
EXECUTE IMMEDIATE lv_stmt;
COMMIT;

EXCEPTION
  WHEN OTHERS THEN
    DBMS_OUTPUT.PUT_LINE('[ADD_TO_WORKLIST - ERROR] Uable to insert record into table: '||WORKLIST_TABLE);
	DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
END;

PROCEDURE STORE_COMPRESS_STATS
  (obj_type IN VARCHAR2, obj_owner IN VARCHAR2, obj_name IN VARCHAR2, obj_partition IN VARCHAR2 DEFAULT NULL, compress_type IN VARCHAR2,
   sample_pct NUMBER DEFAULT 0, prefix_length NUMBER DEFAULT 0, comp_ratio IN NUMBER, org_size IN NUMBER, comp_size IN NUMBER)
AS
  lv_count			NUMBER;
  lv_stmt			VARCHAR2(2000);
BEGIN

-- Make sure the statistics table table exists
SELECT COUNT(1) INTO lv_count FROM user_tables WHERE table_name=STATS_TABLE;
IF (lv_count = 0) THEN
  RAISE_APPLICATION_ERROR(-20001,'[STORE_COMPRESS_STATS - ERROR] Compression statistics table '||STATS_TABLE||' does not exist'); 
END IF;

lv_stmt := 'INSERT INTO '||STATS_TABLE||' (chk_dt,obj_type,obj_owner,obj_name,obj_partition,compress_type, sample_pct, prefix_length, comp_ratio,org_size,comp_size) VALUES (SYSDATE,'''||
obj_type||''','''||obj_owner||''','''||obj_name||''','''||obj_partition||''','''||compress_type||''','||sample_pct||','||prefix_length||','||comp_ratio||','||org_size||','||comp_size||')';
EXECUTE IMMEDIATE lv_stmt;
COMMIT;

EXCEPTION
	WHEN OTHERS THEN
		DBMS_OUTPUT.PUT_LINE('[STORE_COMPRESS_STATS - ERROR] Uable to insert record into table: '||STATS_TABLE);
		DBMS_OUTPUT.PUT_LINE(SUBSTR(SQLERRM, 1, 100));
END;

PROCEDURE SETUP_TABLES
AS
  lv_stmt			VARCHAR2(2000);
BEGIN
  BEGIN
    lv_stmt := 'CREATE TABLE '||WORKLIST_TABLE||' (work_id NUMBER, obj_type VARCHAR2(20), obj_owner VARCHAR2(30), obj_name VARCHAR2(30), obj_partition VARCHAR2(30),'||
	'compress_type VARCHAR2(15), sample_pct NUMBER, sample_ts VARCHAR2(30), status VARCHAR2(20), start_dt DATE, finish_dt DATE, emsg VARCHAR2(200), CONSTRAINT mcl_comp_calc_worklist_pk PRIMARY KEY (work_id))';
    EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[SETUP_TABLES - ERROR] Unable to create worklist table: '||WORKLIST_TABLE);
  END;
  
  BEGIN
    lv_stmt := 'CREATE TABLE '||STATS_TABLE||' (chk_dt DATE, obj_type VARCHAR2(20), obj_owner VARCHAR2(30), obj_name VARCHAR2(30), obj_partition VARCHAR2(30),'||
	'compress_type VARCHAR2(15), sample_pct NUMBER, prefix_length NUMBER, comp_ratio NUMBER, org_size NUMBER, comp_size NUMBER)';
	EXECUTE IMMEDIATE lv_stmt;
  EXCEPTION
    WHEN OTHERS THEN
	  DBMS_OUTPUT.PUT_LINE('[SETUP_TABLES - ERROR] Unable to create statistics table: '||STATS_TABLE);
  END;

EXCEPTION
  WHEN OTHERS THEN
    DBMS_OUTPUT.PUT_LINE('[SETUP_TABLES - ERROR] General error'); 
    RAISE_APPLICATION_ERROR(-20002,'[SETUP_TABLES - ERROR] General error:'||SUBSTR(SQLERRM, 1, 100));
END;

END MCL_COMPRESS_CALC;
/